from pydantic import BaseModel, FilePath, HttpUrl, validator, Field
from dataclasses import dataclass
from typing import List, Dict, Optional, Literal, Union, Any
from collections import Counter
import yaml
import sqlparse
from collections import defaultdict, Counter
from typeguard import typechecked
from itertools import chain
from pathlib import Path
import pandas as pd
import numpy as np
from tidydata.variable import (
    nasum,
    namin,
    namax,
    naprod,
    namean,
    nalog, 
    where, 
    replace, 
    fill,
    ciloc, 
    clast_notna, 
    cfirst_notna,
    cbfillna,
    cffillna,
    reval,
    rassign,
    to_date,
    back1year,
    datecount,
    leftsum)

FUNCTION_MAP = {
    "clast_notna": clast_notna,
    "reval": reval,
    "rassign": rassign,
    "cbfillna": cbfillna
}

global_counter = 0

@typechecked
def _uid(prefix:str = '_uid'):
    global global_counter
    global_counter += 1
    return prefix + str(global_counter)



PandasType = Literal['int', 'float', 'bool', 'str', 'cat', 'date']
ValidValue = Union[int, float, bool, str]



Ext2IOtype = {
            '.csv': 'csv',
            '.tsv': 'tsv',
            '.dta': 'stata',
            '.xls': 'excel',
            '.xlsx': 'excel',
            '.parquet': 'parquet'
}
        

      
PANDAS_DTYPES = {
  'int': 'Int64',
  'uint': 'UInt64',
  'str': 'string',
  'bool': 'boolean',
  'float': 'Float64',
  'cat': 'category',
  'date': 'datetime64[ns]'
}


    

class ColumnBase(BaseModel):
    name: str                           # 选取需要处理的列名
    label: str              # 导出stata时导出为变量标签
    dtype: PandasType       # 转换列的数据类型

class Column(ColumnBase):
    alias: Optional[str]               # 重命名列
    valid_values: Optional[List[List[ValidValue]] | List[ValidValue]]   # 有效值范围
    new_values: Optional[Dict[Any, Any]] # 有效值修改
    value_labels: Optional[Dict[int, str]] # 当列数据类型为整数时， 导出为stata时为值标签
    obs: Optional[int | str]
    expr: Optional[str]

    @validator('dtype')
    def check_and_transform_dtype(cls, v):
        if v is not None:
            if v not in PANDAS_DTYPES:
                raise ValueError(f"dtype '{v}' is not valid. It must be one of {list(PANDAS_DTYPES.keys())}.")
            return PANDAS_DTYPES[v]
        return v

    @validator('valid_values')
    def check_and_transform_valid_values(cls, v):
        if isinstance(v, list):
            if all(isinstance(sub_v, list) for sub_v in v):
                return [tuple(sub_v) if len(sub_v) == 2 and all(isinstance(i, (int, float)) for i in sub_v) else sub_v for sub_v in v]
        return v



    @validator('new_values')
    def check_and_transform_new_values(cls, v):
        if v is not None:
            return {
                (key if key != 'NA' else (None, pd.NA, np.nan)): (value if value != 'NA' else pd.NA)
                for key, value in v.items()
            }
        return v

    
    @validator('value_labels')
    def check_value_labels(cls, v, values):
        if values['dtype'] != PANDAS_DTYPES['int'] and v is not None:
            raise ValueError(f"value_labels in column '{values['name']}' must be None when dtype is not 'int'")
        return v
    

    @validator('expr')
    def transform_expr_to_lambda(cls, v):
        if v is not None:
            return  eval('lambda x: ' + v)
        return v

class ColumnGroup(BaseModel):
    gname: str
    members: List[Column]
    glabel: Optional[str]
    galias: Optional[str]
    gdtype: Optional[PandasType]
    gvlabel: Optional[Dict[int, str]]
    obsname: Optional[str]




    @validator('gdtype')
    def check_and_transform_gdtype(cls, v):
        if v is not None:
            if v not in PANDAS_DTYPES:
                raise ValueError(f"gdtype '{v}' is not valid. It must be one of {list(PANDAS_DTYPES.keys())}.")
            return PANDAS_DTYPES[v]
        return v

    @validator('gvlabel')
    def check_group_value_labels(cls, v, values):
        if values['gdtype'] != PANDAS_DTYPES['int'] and v is not None:
            raise ValueError(f"value_labels in column group '{values['gname']}' must be None when its dtype is not 'int'")
        return v
    
    @validator('members')
    def check_obs_unique(cls, v, values):
        obs_values = {member.obs for member in v}
        if len(obs_values) != len(v):
            raise ValueError(f"obs values in members of group {values['gname']} must be unique.")
        return v
    
class Dataset(BaseModel):
    id: str
    description: Optional[str]
    location: FilePath | HttpUrl
    rows: Optional[List[str]]
    newcols: Optional[List[Column]]
    idcols: List[Column]
    longcols: Optional[List[Column]]
    widecols: Optional[List[ColumnGroup]]

    @validator('location')
    def stringify_location(cls, v):
        return str(v)
    
    @property
    def reader_type(self) -> str:
        """创建文件读取类型的只读属性"""
        ext = Path(self.location).suffix
        return Ext2IOtype[ext]
        
    @property
    def id_names(self):
        """数据集的ID列名，用于检查数据集ID是否唯一, 宽转长和横向合并"""
        return [col.name for col in self.idcols]

    @property
    def group_names(self):
        if self.widecols:
            group_names = defaultdict(list)

            for colgroup in self.widecols:
                group_names[colgroup.obsname].append(colgroup.gname)
                
            return dict(group_names)

    @property
    def widename_mapper(self):
        if self.widecols:
            return {
                member.name: f"{colgroup.gname}-{member.obs}"
                for colgroup in self.widecols 
                for member in colgroup.members 
                if member.name != f"{colgroup.gname}-{member.obs}"
            }

    @property
    def _pre_columns(self):
        return chain(
            self.idcols, 
            self.longcols or [], 
            *(g.members for g in self.widecols or [])
        )

    @property
    def pre_column_names(self):
        return [col.name for col in self._pre_columns]

    @property
    def pre_column_dtypes(self):
        # 使用 all_columns 属性
        column_dtypes = {col.name: col.dtype for col in self._pre_columns if col.dtype}

        return column_dtypes if column_dtypes else None
    
    @property
    def _post_columns(self):
        return chain(
            self.idcols, 
            self.longcols or [], 
            self.widecols or []
        )
        
    @property
    def post_column_aliases(self):
        column_aliases = defaultdict(str)

        for col in self._post_columns:
            if isinstance(col, Column) and col.alias and col.alias != col.name:
                column_aliases[col.name] = col.alias
            elif isinstance(col, ColumnGroup) and col.galias and col.galias != col.gname:
                column_aliases[col.gname] = col.galias

        return dict(column_aliases) if column_aliases else None

    @property
    def post_column_value_labels(self):
        column_value_labels = defaultdict(str)

        for col in self._post_columns:
            if isinstance(col, Column) and col.valid_values:
                column_value_labels[col.name] = col.valid_values
            elif isinstance(col, ColumnGroup) and col.gvlabel:
                column_value_labels[col.name] = col.gvlabel

        return dict(column_value_labels) if column_value_labels else None

    @property
    def pre_column_valid_values(self):
        
        valid_values = {col.name: col.valid_values for col in self._pre_columns if col.valid_values}

        return valid_values if valid_values else None

    @property
    def pre_column_new_values(self):

        new_values = {col.name: col.new_values for col in self._pre_columns if col.new_values}

        return new_values if new_values else None

    @property
    def new_column_dtypes(self):
        if self.newcols:
            new_column_dtypes = {col.name: col.dtype for col in self.newcols if col.dtype}
            return new_column_dtypes if new_column_dtypes else None

    @property
    def new_column_exprs(self):
        if self.newcols:
            new_column_exprs = [{col.name: col.expr} for col in self.newcols if col.expr is not None]
            return new_column_exprs if new_column_exprs else None



class Query(BaseModel):
    """构造单个原始表的查询语句"""
    from_: str = Field(..., alias='from')
    select: Optional[List[str]]
    where: Optional[str]
    groupby: Optional[str]
    having: Optional[str]
    orderby: Optional[str]

    @property
    def sql(self):
        clauses = [
            f"SELECT {', '.join(self.select) if self.select else '*'}",
            f"FROM {self.from_}",
        ]
        if self.where:
            clauses.append(f"WHERE {self.where}")
        if self.groupby:
            clauses.append(f"GROUP BY {self.groupby}")
        if self.having:
            if self.groupby is None:
                raise ValueError("SQL Error: HAVING can only be used when GROUP BY is specified.")
            clauses.append(f"HAVING {self.having}")
        if self.orderby:
            clauses.append(f"ORDER BY {self.orderby}")

        return sqlparse.format(" ".join(clauses), reindent=True, keyword_case='upper')



class Edit(BaseModel):
    """构造合并表的单个编辑函数"""
    func: str = Field(..., description="Function to be applied",
                      choices=list(FUNCTION_MAP.keys()))  # 使用函数名列表作为选择
    kwargs: Optional[Dict[str, Any]]

    @validator('func')
    def map_func(cls, v):
        return FUNCTION_MAP[v]
    
    @validator('kwargs')
    def map_kwargs(cls,v):
        return v if v else {}
    
    @property
    def editor(self):
        return {'func':self.func, 'kwargs': self.kwargs}


    
@typechecked
def union(sql_statements: List[str], method: str, use: Optional[List[str]] = None, where: Optional[str]=None) -> str:
    """纵向合并表的SQL语句解析"""
    columns = '*' if not use else ', '.join(use)
    queries = [stmt.replace('SELECT *', f'SELECT {columns}') for stmt in sql_statements]
    unioned_result = f"\n{method}\n  ".join(f"{query}" for query in queries)
    return unioned_result if not where else f"SELECT * FROM ({unioned_result}) WHERE {where}"


@typechecked
def join(sql_statements: List[str], method: str, use: List[str], where: Optional[str]=None) -> str:
    # For join operations, using clause is necessary
    if not use:
        raise ValueError('The "use" argument is required for JOIN operations')
    using_clause = ', '.join(use)
    # Add a UUID alias to the first query
    
    join_query = f"({sql_statements[0]}) AS {_uid()}"
    # And then add a UUID alias for each additional query, and join them
    for stmt in sql_statements[1:]:
        join_query += f"\n{method}\n({stmt}) AS {_uid()} USING ({using_clause})"
    
    joined_result = f"SELECT * FROM ({join_query})" if not where else f"SELECT * FROM ({join_query}) WHERE {where}"
    return  joined_result



@typechecked
def combine(sql_statements: List[str], method: str, use: Optional[List[str]] = None, where: Optional[str]=None) -> str:
    if len(sql_statements) < 2:
        raise ValueError("At least two SQL statements are required for combining.")
    
    
    COMBINE_METHOD = {
        'union': 'UNION',
        'union_all': 'UNION ALL',
        'left_join': 'LEFT JOIN',
        'right_join': 'RIGHT JOIN',
        'join': 'JOIN',
        'full_join': 'FULL JOIN'
    }
    
    
    if method in {'union', 'union_all'}:
        return union(sql_statements, COMBINE_METHOD[method], use, where)

    elif method in {'left_join', 'right_join', 'join', 'full_join'}:
        return join(sql_statements, COMBINE_METHOD[method], use, where)
    

    else:
        raise ValueError(f"Unknown method: {method}")


class Combine(BaseModel):
    method: Literal['union','union_all','left_join','right_join', 'join','full_join']
    use: List[str]
    where: Optional[str]

class FromAction(BaseModel):
    fid: Optional[str]
    description: Optional[str]
    query: List[Query] 
    combine: Optional[Combine]
    edit: Optional[List[Edit]]


    @validator('combine')
    def check_combine(cls, combine, values):
        query = values.get('query')
        if query is not None:
            if combine is None and len(query) != 1:
                raise ValueError('When "combine" is None, "query" must have exactly 1 element.')
            elif combine is not None and len(query) <= 1:
                raise ValueError('When "combine" is not None, "query" must have more than 1 element.')
        return combine



    @property
    def query_action(self):
        if len(self.query) == 1:
            sql = sqlparse.format(self.query[0].sql, reindent=True, keyword_case='upper')
        else:
            sql_statements = [q.sql for q in self.query]
            method = self.combine.method
            use = self.combine.use
            where = self.combine.where
            sql = combine(sql_statements, method=method, use=use, where=where)

        return sql
    
    @property
    def edit_action(self):
        if self.edit:
            return [edit.editor for edit in self.edit]    
    

class ToAction(BaseModel):
    combine: Optional[Combine]
    edit: Optional[List[Edit]]
    export: Optional[Literal['parquet','stata']] = None

    @property
    def edit_action(self):
        if self.edit:
            return [edit.editor for edit in self.edit] 
        
        
class Export(BaseModel):
    name: str 
    description: str
    from_tables: List[FromAction]
    to_table: Optional[ToAction]

    @validator('to_table')
    def validate_tables(cls, v, values):
        from_tables = values.get('from_tables')
        if (not v or not v.combine) and len(from_tables) >= 2:
            raise ValueError('When from_tables contain more than one element, the combine attribute of to_table must be not None')
        elif v and v.combine and len(from_tables) == 1:
            raise ValueError('When from_tables contain only one element, the combine attribute of to_table must be None')
        return v

    @property
    def from_actions(self) -> List[Dict[str, Any]]:
        from_actions = [{'fid': fa.fid if fa.fid else '_'+self.name+str(i+1),
                         'description': fa.description,
                         'query_action': fa.query_action, 
                         'edit_action': fa.edit_action} for i, fa in enumerate(self.from_tables)]
        return from_actions


    @property
    def to_actions(self):
        
        sql_statements = [f"SELECT * FROM {fa_dict['fid']}" for fa_dict in self.from_actions]
        query_action = (
                combine(sql_statements = sql_statements, 
                        method = self.to_table.combine.method, 
                        use = self.to_table.combine.use) 
                if self.to_table and self.to_table.combine 
                else sql_statements[0]
            )
        
        edit_action = self.to_table.edit_action if self.to_table else None
        
        ToFileExtension = {
            'stata':'.dta',
            'parquet':'.parquet'
        }
        
        to_file = {'filetype':self.to_table.export, 'filename': self.name+ToFileExtension[self.to_table.export]} if self.to_table and self.to_table.export else None
  
        return {'tid':self.name, 'description': self.description, 'to_file': to_file, 'query_action': query_action, 'edit_action': edit_action}


class DataBase(BaseModel):
    fast_load: Optional[str]=None
    source: str  = 'source.duckdb'
    export: str  = './'

    @validator('fast_load',always=True)
    def check_fast_load(cls, v):
        if v and not v.endswith('.duckdb'):
            raise ValueError(f"Database for fast load {v} should be a file path with a suffix '.duckdb'")
        return Path(v) if v else v
        
    @validator('source',always=True)
    def check_source(cls, v):
        if not v.endswith('.duckdb'):
            raise ValueError(f"Database for source {v} should be a file path with a suffix '.duckdb'")
        return Path(v)
    

    @validator('export',always=True)
    def check_export(cls, v):
        if not v.endswith('/'):
            raise ValueError(f"Database for export {v} should be a directory with a suffix '/'")
        
        path = Path(v)
        if not path.exists():
            path.mkdir(parents=True, exist_ok=True)
        
        if not path.is_dir():
            raise ValueError(f"Database for export {v} should be a directory")
        
        return path

class Config(BaseModel):
    database: DataBase = DataBase()
    sources: List[Dataset]
    exports: Optional[List[Export]]


    
    @validator('sources')
    def validate_unique_ids(cls, sources):
        ids = [source.id for source in sources]
        id_counts = Counter(ids)
        duplicate_ids = [id for id, count in id_counts.items() if count > 1]
        if duplicate_ids:
            raise ValueError(f"The id(s) {duplicate_ids} of Dataset in sources are not unique.")
        return sources

    @classmethod
    def from_yaml(cls, path: str):
        with open(path, 'r') as f:
            data = yaml.safe_load(f)
        return cls(**data)


import click
import shutil
from pathlib import Path
from tidydata.config import Config  # Assumption that you have a Config class in config module of your package
from tidydata.core import TidySource, TidyExport
import os

@click.group()
def tidydata():
    pass

@tidydata.command()
@click.argument('name')
def new(name):
    """Creates a new project with the given name."""
    base_dir = Path(name)
    base_dir.mkdir(parents=True, exist_ok=True)
    (base_dir / ".data/raw").mkdir(parents=True, exist_ok=True)
    (base_dir / ".data/preload").mkdir(parents=True, exist_ok=True)

    yaml_file = base_dir / "data.yaml"
    with yaml_file.open("w") as f:
        f.write("sources:")

@tidydata.command()
@click.argument('name')
def remove(name):
    """Removes an existing project with the given name."""
    base_dir = Path(name)
    if base_dir.exists():
        shutil.rmtree(base_dir)
    else:
        click.echo(f"The project '{name}' does not exist.")

@tidydata.command()
@click.option('-f', '--from-yaml', default='data.yaml', help='Name of the YAML file to process.')
@click.option('-t','--to-dir', default=None, help='Specify the export directory.')
@click.option('-p','--project-dir', default=None, help='Specify the project working directory.')
def run(from_yaml, to_dir, project_dir):
    """Processes the specified YAML file and creates a Config object from it."""
    
    original_dir = os.getcwd()  # Save the original working directory

    # If a project directory is specified, change to it
    if project_dir:
        os.chdir(project_dir)

    yaml_file = Path(from_yaml)
    if not yaml_file.exists():
        click.echo(f"The file '{from_yaml}' does not exist in the current directory.")
    else:
        config = Config.from_yaml(yaml_file)
        TidySource(config)
        TidyExport(config, export_dir=to_dir)
        click.echo(f"Data in '{from_yaml}' was successfully cleaned, and are saved to {to_dir}")

    os.chdir(original_dir)  



if __name__ == "__main__":
    tidydata()



# 导入数据集
import pandas as pd
import numpy as np
import duckdb
from pathlib import Path
from typeguard import typechecked
from typing import List, Optional, Union, Dict, Tuple, Literal, Any
from collections import defaultdict
from tidydata.config import Config
import logging
from rich import print
import re
from tidydata.variable import (
    ciloc, clast_notna, cfirst_notna, reval
)


logging.basicConfig(level=logging.INFO)

Number = Union[int, float]
Scalar = Union[str, int, float, bool]
ScalarNA = Union[str, int, float, bool, pd.NA, np.nan, None, pd.NaT] 



PANDAS_TO_NUMPY = {
    pd.BooleanDtype(): np.bool_,
    pd.Int8Dtype(): np.int8,
    pd.Int16Dtype(): np.int16,
    pd.Int32Dtype(): np.int32,
    pd.Int64Dtype(): np.int64,
    pd.UInt8Dtype(): np.uint8,
    pd.UInt16Dtype(): np.uint16,
    pd.UInt32Dtype(): np.uint32,
    pd.UInt64Dtype(): np.uint64,
    pd.Float32Dtype(): np.float32,
    pd.Float64Dtype(): np.float64,
    pd.StringDtype(): np.str_
}

STATA_VERSION = {
    'Stata10': 114,
    'Stata13': 117,
    'Stata14': 118,
    'Stata14Max': 119
}






@typechecked
class Table:
    """Table类"""
  
    def __init__(self, name: str, df: pd.DataFrame):
        self.df = df
        self.name = name
        self._check_mixed_dtypes()


    def _check_mixed_dtypes(self):
        """检查并警告混合数据类型"""
        mixed_type_cols = []
        mixed_integer_float_cols = []
        mixed_integer_cols = []
        unknown_array_cols = []

        for col in self.df.columns:
            inferred_dtype = pd.api.types.infer_dtype(self.df[col])
            if inferred_dtype == 'mixed':
                mixed_type_cols.append(col)
            elif inferred_dtype == 'mixed-integer-float':
                mixed_integer_float_cols.append(col)
            elif inferred_dtype == 'mixed-integer':
                mixed_integer_cols.append(col)
            elif inferred_dtype == 'unknown-array':
                unknown_array_cols.append(col)

        if len(mixed_type_cols) > 0:
            logging.warning(f"The dataframe contains columns of type 'mixed': {mixed_type_cols}. "
                            "It is recommended to use the 'replace_values' method to replace non-null values with a consistent type.")
        if len(mixed_integer_float_cols) > 0:
            logging.warning(f"The dataframe contains columns of type 'mixed-integer-float': {mixed_integer_float_cols}. "
                            "It is recommended to use the 'replace_values' method to replace non-null values with a consistent type.")
        if len(mixed_integer_cols) > 0:
            logging.warning(f"The dataframe contains columns of type 'mixed-integer': {mixed_integer_cols}. "
                            "It is recommended to use the 'replace_values' method to replace non-null values with a consistent type.")
        if len(unknown_array_cols) > 0:
            logging.warning(f"The dataframe contains columns of type 'unknown-array': {unknown_array_cols}. "
                            "It is recommended to use the 'replace_values' method to replace non-null values with a consistent type.")
            
    @classmethod
    def from_stata(cls, name: str, file: Path | str, usecols: Optional[List[str]] = None, convert_categoricals: bool = False, fast_load_path: Optional[str | Path] = None):
        tbname = Path(file).stem
        if fast_load_path:
            fast_load_path = str(fast_load_path)
            with duckdb.connect(fast_load_path) as conn: 
                try:
                    columns = ", ".join(usecols) if usecols else "*"
                    df = conn.execute(f"SELECT {columns} FROM {tbname};").df()
                except Exception as e:
                    logging.warning(f"正在导入文件到DuckDB'{fast_load_path}'为表{tbname}")
                    df = pd.read_stata(file, convert_categoricals=convert_categoricals)
                    conn.execute(f"CREATE TABLE {tbname} AS SELECT * FROM df")

                    df = conn.execute(f"SELECT {columns} FROM {tbname};").df()
        else:
            df = pd.read_stata(file, columns=usecols, convert_categoricals=convert_categoricals)
            
        return cls(name, df.convert_dtypes())



    @classmethod
    def from_duckdb(cls, name: str, connect: Path | str, sql_statement: str):
    
        with duckdb.connect(str(connect)) as conn:
            return cls(name, conn.sql(sql_statement).df().convert_dtypes())



    @classmethod
    def from_csv(cls, name: str, file: Path | str, usecols: Optional[List[str]] = None, fast_load_path: Optional[str| Path] = None):
        
        tbname = Path(file).stem
        if fast_load_path:
            fast_load_path = str(fast_load_path)
            with duckdb.connect(fast_load_path) as conn:
                try:
                    columns = ", ".join(usecols) if usecols else "*"
                    df = conn.execute(f"SELECT {columns} FROM {tbname};").df()
                except Exception as e:
                    logging.info(f"正在导入文件到DuckDB'{fast_load_path}'为表{tbname}")
                    conn.execute("INSTALL httpfs;")
                    conn.execute(f"CREATE TABLE {tbname} AS SELECT * FROM read_csv_auto('{file}');")
                    df = conn.execute(f"SELECT {columns} FROM {tbname};").df()
        else:
            df = pd.read_csv(file, usecols=usecols, sep=',')
        return cls(name, df.convert_dtypes())


    @classmethod
    def from_tsv(cls, name: str, file: Path | str, usecols: Optional[List[str]] = None, fast_load_path: Optional[str|Path] = None):
        
        tbname = Path(file).stem
        if fast_load_path:
            fast_load_path = str(fast_load_path)
            with duckdb.connect(fast_load_path) as conn:
                try:
                    columns = ", ".join(usecols) if usecols else "*"
                    df = conn.execute(f"SELECT {columns} FROM {tbname};").df()
                except Exception as e:
                    logging.info(f"正在导入文件到DuckDB'{fast_load_path}'为表{tbname}")
                    conn.execute("INSTALL httpfs;")
                    conn.execute(f"CREATE TABLE {tbname} AS SELECT * FROM read_csv_auto('{str(file)}')")
                    df = conn.execute(f"SELECT {columns} FROM {tbname};").df()
        else:
            df = pd.read_csv(file, usecols=usecols, sep='\t')
        return cls(name, df.convert_dtypes())

    @classmethod
    def from_excel(cls, name: str, file: Path | str, usecols: Optional[List[str]] = None,  sheet_name: str = 'Sheet1', fast_load_path: Optional[str|Path] = None):
        tbname = Path(file).stem
        if fast_load_path:
            fast_load_path = str(fast_load_path)
            with duckdb.connect(fast_load_path) as conn:
                table_exists = conn.execute(f"SELECT name FROM tables WHERE name='{tbname}';").fetchone() is not None
                
                if not table_exists:
                    df = pd.read_excel(file, sheet_name=sheet_name)
                    conn.execute(f"CREATE TABLE {tbname} AS", df)

                columns = ", ".join(usecols) if usecols is not None else "*"
                df = conn.execute(f"SELECT {columns} FROM {tbname};").fetch_df()
        else:
            df = pd.read_excel(file, usecols=usecols, sheet_name=sheet_name)

        return cls(name, df.convert_dtypes())


    @classmethod
    def from_parquet(cls, name: str, file: Path | str, usecols: Optional[List[str]] = None, fast_load_path: Optional[str | Path] = None):
        tbname = Path(file).stem
        
        if fast_load_path:
            fast_load_path = str(fast_load_path)
            with duckdb.connect(fast_load_path) as conn:
                try:
                    columns = ", ".join(usecols) if usecols else "*"
                    df = conn.execute(f"SELECT {columns} FROM {tbname};").df()
                except Exception as e:
                    logging.info(f"正在导入文件到DuckDB'{fast_load_path}'为表{tbname}")
                    conn.execute("INSTALL httpfs;")
                    
                    conn.execute(f"CREATE TABLE {tbname} AS SELECT * FROM read_parquet('{str(file)}')")
                    df = conn.execute(f"SELECT {columns} FROM {tbname};").df()
        else:
            df = pd.read_parquet(file, columns=usecols, engine='pyarrow')

        return cls(name, df.convert_dtypes())



    def select_columns(self, cols: List[str]):
        filtered_df = self.df.filter(items=cols)
        return self.__class__(self.name, filtered_df)

    def select_rows(self, exprs: List[str]):
        exprs = [f'{expr.replace(" is not NA", " == " + expr.split()[0])}' for expr in exprs]
        query_expr = " & ".join(exprs)
        queried_df = self.df.query(query_expr)
        return self.__class__(self.name, queried_df)


    def downcast_pandas_dtypes(self, nocats: Optional[List[str]]=None):
        df = self.df.convert_dtypes()

        if nocats is None:
            nocats = []

        number_mapping = {
            'int': {
                'Int8': (np.iinfo(np.int8).min, np.iinfo(np.int8).max),
                'Int16': (np.iinfo(np.int16).min, np.iinfo(np.int16).max),
                'Int32': (np.iinfo(np.int32).min, np.iinfo(np.int32).max),
            },
            'uint': {
                'UInt8': (0, np.iinfo(np.uint8).max),
                'UInt16': (0, np.iinfo(np.uint16).max),
                'UInt32': (0, np.iinfo(np.uint32).max),
            },
        }

        dtype_mapping = defaultdict(dict)

        for col in df.columns:
            if pd.api.types.is_integer_dtype(df[col]):
                min_val, max_val = df[col].dropna().min(), df[col].dropna().max()
                dtype_class = 'int' if pd.api.types.is_signed_integer_dtype(df[col]) else 'uint'
                for dtype_type, (min_int, max_int) in number_mapping[dtype_class].items():
                    if min_val >= min_int and max_val <= max_int:
                        dtype_mapping[col] = dtype_type
                        break

            elif pd.api.types.is_float_dtype(df[col]):
                min_val, max_val = df[col].dropna().min(), df[col].dropna().max()
                if min_val >= np.finfo(np.float32).min and max_val <= np.finfo(np.float32).max:
                    dtype_mapping[col] = 'Float32'

            elif pd.api.types.is_string_dtype(df[col]) and col not in nocats:
                dtype_mapping[col] = 'category'

        if dtype_mapping:
            df = df.astype(dict(dtype_mapping))

        return self.__class__(self.name, df)



    def _check_float_to_int(self, dtype_mapper):

        df = self.df.convert_dtypes()
        for col, target_dtype in dtype_mapper.items():
            if pd.api.types.is_float_dtype(df[col]) and ('int' in target_dtype or 'uint' in target_dtype):
                logging.warning(f"Converting from float to integer for column '{col}' may cause data loss. "
                                "Please check your dtype_mapper and confirm.")
        return self.__class__(self.name, df)


    def convert_dtypes(self, dtype_mapper: Dict[str, str], downcast: bool=False, nocats: Optional[List[str]]=None):

        if not nocats:
            nocats = [key for key, value in dtype_mapper.items() if value == 'string']
    
        df = (self._check_float_to_int(dtype_mapper)
            .df.convert_dtypes()
            .astype(dtype_mapper))

        return self.__class__(self.name, df)

    def specify_valid_values(self, valid_values: Dict[str, List[Union[Scalar, Tuple[Number, Number]]]]):

        df = self.df.copy()

        for col, values in valid_values.items():
            if col in df.columns:
                scalar_values = [value for value in values if not isinstance(value, tuple)]
                tuple_values = [value for value in values if isinstance(value, tuple)]

                # Using mask for scalar_values
                if scalar_values:
                    df[col] = df[col].mask(~df[col].isin(scalar_values))

                # Using where for tuple_values
                for value_range in tuple_values:
                    min_val, max_val = value_range
                    df[col] = df[col].where(df[col].between(min_val, max_val).fillna(False), other=pd.NA)

        return self.__class__(self.name, df)



    def replace_values(self, value_mapper: Dict[str, Dict[ScalarNA | Tuple[ScalarNA], ScalarNA]] | Dict[ScalarNA | Tuple[ScalarNA], ScalarNA]):
        
        if not all(isinstance(val, dict) for val in value_mapper.values()):
            value_mapper = {col: value_mapper for col in self.df.columns}

        replace_dict = {}
        for col, mapper in value_mapper.items():
            replace_dict[col] = {k: v for key, v in mapper.items() for k in (key if isinstance(key, tuple) else [key])}

        replaced_df = self.df.replace(replace_dict)

        return self.__class__(self.name, replaced_df)

    def rename_columns(self, column_mapper: Dict[str, str]):
        renamed_df = self.df.rename(columns=column_mapper)
        return self.__class__(self.name, renamed_df)

    def add_columns(self, column_exprs: List[Dict[str, callable]]):
        new_df = self.df.copy().convert_dtypes() 
        for expr in column_exprs:
            new_df = new_df.assign(**expr).convert_dtypes()

        return self.__class__(self.name, new_df)

    def apply_functions(self, functions: List[Dict[str, Any]]) -> 'Table':
        
        df = self.df.copy()
        
        for func_dict in functions:
            df = df.pipe(func_dict['func'], **func_dict['kwargs'])

        return Table(self.name, df)
        

    def melt_columns(self, i: List[str], j: Dict[str, List[str]], prename_mapper: Optional[Dict[str, str]]=None, sep: str = '-'):
        df = self.df.copy()

        if prename_mapper:
            df.rename(columns=prename_mapper, inplace=True)

        for j_val, stubnames in j.items():
            df = pd.wide_to_long(df, stubnames, i=i, j=j_val, sep=sep, suffix=r'\d+').reset_index()
            
            df.loc[df[stubnames].isna().all(axis=1), j_val] = pd.NA
            df.drop_duplicates(subset=i+[j_val], inplace=True)
            
            subset_columns = df.columns.difference(i + [j_val])
            df.dropna(subset=subset_columns, how='all', inplace=True)
            
            df[j_val] = df[j_val].astype('Int64')
        return self.__class__(self.name, df)


    def _to_numpy_dtypes(self):
        """转换 pandas 扩展类型到相应的 numpy 类型"""
        df = self.df.copy()

        # 对于每一列
        for col, dtype in df.dtypes.items():
            # 如果列中有缺失值，则转为 'object'
            if df[col].isnull().any():
                df[col] = df[col].astype('object').fillna(np.nan)
            # 否则按照 PANDAS_TO_NUMPY 映射进行转换
            elif dtype in PANDAS_TO_NUMPY:
                df[col] = df[col].astype(PANDAS_TO_NUMPY[dtype])

        # 返回新的 TidyData 对象
        return self.__class__(self.name, df)

    def to_stata(self, 
                to_path: Path | str, 
                file_label: Optional[str]=None, 
                columns: Optional[List[str]]=None, 
                column_labels: Optional[Dict[str, str]]=None, 
                value_labels: Optional[Dict[str, Dict[int, str]]]=None,
                version: Literal['Stata10','Stata13','Stata14', 'Stata14Max'] = 'Stata14'):
        
        export_df = (self.select_columns(columns) if columns else self)._to_numpy_dtypes().df
        
        export_df.to_stata(to_path, write_index=False, version=STATA_VERSION[version], 
                        data_label=file_label, variable_labels=column_labels, value_labels=value_labels)
        
    def to_parquet(self, to_path: Path | str, columns: Optional[List[str]]= None, compression: bool = True):
        
        export_df = (self.select_columns(columns) if columns else self).df

        # 导出数据
        export_df.to_parquet(
            to_path,
            engine='pyarrow',
            compression='gzip' if compression else None
        )
        
    def to_duckdb(self, to_path: str | Path, if_exists: str = 'fail'):
        con = duckdb.connect(str(to_path))

        # Register the DataFrame with DuckDB
        con.register('df', self.df.sort_index(axis=1))

        # 检查表是否存在
        query = f"""
        SELECT count(*) 
        FROM information_schema.tables 
        WHERE table_name = '{self.name}'
        """
        table_exists = con.execute(query).fetchone()[0] > 0

        # 根据 if_exists 选项执行相应的操作
        if table_exists:
            if if_exists == 'fail':
                raise ValueError(f"Table '{self.name}' already exists.")
            elif if_exists == 'replace':
                con.execute(f"DROP TABLE {self.name}")
            elif if_exists == 'append':
                pass
            else:
                raise ValueError(f"Invalid value for 'if_exists': {if_exists}")
        
        # 将 DataFrame 导入到 DuckDB 数据库
        con.execute(f"CREATE TABLE {self.name} AS SELECT * FROM df")

        # 关闭数据库连接
        con.close()



@typechecked
class TidySource:
    """TidySource类"""
    
    def __init__(self, config: Config) -> None:
        self.sources  = config.sources
        self.fast_load_db = config.database.fast_load
        self.source_db = config.database.source
        
        if not self.sources:
            logging.error('There is no entity in sources')
        
        self.load()
        
        logging.info(f"完成数据预处理.")
    def load(self):
        for ds in self.sources:
            match ds.reader_type:
                case 'stata':
                    self.load_and_process_data(Table.from_stata, ds)
                case 'csv':
                    self.load_and_process_data(Table.from_csv, ds)
                case 'tsv':
                    self.load_and_process_data(Table.from_tsv, ds)
                case 'excel':
                    self.load_and_process_data(Table.from_excel, ds)
                case 'parquet':
                    self.load_and_process_data(Table.from_parquet, ds)

    def load_and_process_data(self, reader, ds):
        
        logging.info(f"正在导入并预处理数据集{ds.id}...")
        table = reader(ds.id, ds.location, usecols= ds.pre_column_names, fast_load_path=self.fast_load_db)

        if ds.pre_column_new_values:
            logging.info(f"正在值替换...")
            table = table.replace_values(value_mapper= ds.pre_column_new_values)
        
        if ds.pre_column_valid_values:
            logging.info(f"正在设置有效值...")
            table = table.specify_valid_values(valid_values= ds.pre_column_valid_values)
        
        if ds.pre_column_dtypes:
            logging.info(f"正在转换用户指定的列数据类型...")
            table = table.convert_dtypes(dtype_mapper=ds.pre_column_dtypes)
        
        if ds.id_names and ds.group_names and ds.widename_mapper:
            logging.info(f"正在宽格式转长格式...")
            table = table.melt_columns(i=ds.id_names, j=ds.group_names, prename_mapper=ds.widename_mapper)
        
        if ds.post_column_aliases:
            logging.info(f"正在重命名长格式列名...")
            table = table.rename_columns(column_mapper=ds.post_column_aliases)


        if ds.new_column_exprs:
            logging.info(f"正在创建新列...")
            table = table.add_columns(column_exprs=ds.new_column_exprs)
            if ds.new_column_dtypes:
                table = table.convert_dtypes(dtype_mapper=ds.new_column_dtypes)
                
        if ds.rows:
            logging.info(f"正在选取符合要求的行...")
            table = table.select_rows(exprs=ds.rows)
                
        logging.info(f"正在导出到'{self.source_db}'...")
        table.to_duckdb(to_path=self.source_db, if_exists='replace')
        
        logging.info(f"已完成预处理数据集{ds.id}.")
        return table
    

@typechecked
class TidyExport:
    """TidyExport类"""

    def __init__(self, config: Config, export_dir: Optional[str] = None) -> None:
        self.exports = config.exports
        self.db = str(config.database.source)
        self.export_dir = export_dir if export_dir is not None else str(config.database.export)

        if not self.exports:
            logging.error('There is no entity in exports')
                
        logging.info(f"运行导出模块...")
        self.export()

        if Path(self.db).exists():
            Path(self.db).unlink()
        logging.info(f"删除临时数据库...") 
        logging.info(f"完成数据导出.")
        
    def export(self):
        for ds in self.exports:
            for fa in ds.from_actions:

                from_edit = fa['edit_action']
            
                logging.info(f"正在创建Table对象'from_table'...")
                from_table = Table.from_duckdb(name = fa['fid'], connect=self.db, sql_statement=fa['query_action'])
                if from_edit:
                    logging.info(f"正在为Table对象'from_table'应用编辑函数")
                    from_table = from_table.apply_functions(functions=from_edit)
                    
                logging.info(f"完成编辑并以表名{fa['fid']}导入到数据库{self.db}(如果存在则替换!)")
                from_table.to_duckdb(to_path=self.db,if_exists='replace')
                logging.info(f"已完成表{fa['fid']}创建到数据库{self.db}")

            to_file = ds.to_actions['to_file']
            to_edit = ds.to_actions['edit_action']
            
            logging.info(f"正在创建Table对象'to_table'...")
            to_table = Table.from_duckdb(name=ds.to_actions['tid'],connect=self.db, sql_statement=ds.to_actions['query_action'])
            if to_edit:
                logging.info(f"正在为Table对象'to_table'应用编辑函数")
                to_table = to_table.apply_functions(functions=to_edit)
            
            logging.info(f"完成编辑并以表名{ds.to_actions['tid']}导入到数据库{self.db}(如果存在则替换!)")    
            to_table.to_duckdb(to_path=self.db,if_exists='replace')
            logging.info(f"已完成表{ds.to_actions['tid']}创建到数据库{self.db}")
            if to_file:
                logging.info(f"正在导出表{ds.to_actions['tid']}为文件{self.export_dir+'/'+to_file['filename']}")
                match to_file['filetype']:
                    case 'stata':
                        to_table.to_stata(to_path=self.export_dir+'/'+to_file['filename'])
                        logging.info(f"完成导出表{ds.to_actions['tid']}为文件{self.export_dir+'/'+to_file['filename']}")
                    case 'parquet':
                        to_table.to_parquet(to_path=self.export_dir+'/'+to_file['filename'])
                        logging.info(f"完成导出表{ds.to_actions['tid']}为文件{self.export_dir+'/'+to_file['filename']}")



           

        
        
        
        
        
        
        


            



import pandas as pd
import numpy as np
import re
from typing import List,Literal,Optional,Callable, Dict,Tuple
from pandas.api.types import is_numeric_dtype, is_string_dtype, pandas_dtype, is_categorical_dtype
import functools

def datecount(data: Tuple[pd.Series, pd.Series], refs: Optional[Tuple[pd.Series, pd.Series]]=None, 
              unit: Literal['year','quarter','month','week','day']='day', ignore_interval_error: bool=True)->pd.Series:
    data_start, data_end = data
    ref_start, ref_end = refs if refs else data

    df = pd.DataFrame({
        'data_start': data_start,
        'data_end': data_end,
        'ref_start': ref_start,
        'ref_end': ref_end
    }).astype('datetime64[ns]')

    interval_errors = (df['data_start'] > df['data_end']) | (df['ref_start'] > df['ref_end'])
    
    if interval_errors.any():
        if ignore_interval_error:
            df[interval_errors] = pd.NaT
        else:
            raise ValueError("Start time cannot be after end time.")

    df['overlap_start'] = df[['data_start', 'ref_start']].max(axis=1,skipna=False)
    df['overlap_end'] = df[['data_end', 'ref_end']].min(axis=1,skipna=False)
    
    df.loc[df['overlap_end'] < df['overlap_start'], 'overlap_end'] = df['overlap_start']



    if unit == 'year':
        result = df['overlap_end'].dt.year - df['overlap_start'].dt.year
    elif unit == 'quarter':
        result = (df['overlap_end'].dt.year - df['overlap_start'].dt.year) * 4 + df['overlap_end'].dt.quarter - df['overlap_start'].dt.quarter
    elif unit == 'month':
        result = (df['overlap_end'].dt.year - df['overlap_start'].dt.year) * 12 + df['overlap_end'].dt.month - df['overlap_start'].dt.month
    elif unit == 'week':
        result = (df['overlap_end'] - df['overlap_start']).dt.days // 7
    else:  # day
        result = (df['overlap_end'] - df['overlap_start']).dt.days

    return result



def to_date(year, month, day=None):

    day = day if day is not None else pd.Series([1]*len(year))
    
    mask_missing = (year.isnull()) | (month.isnull()) | (day.isnull())
    
    date_series = pd.Series(pd.NaT, index=year.index)
    
    date_series[~mask_missing] = pd.to_datetime(pd.DataFrame({'year': year[~mask_missing], 
                                                              'month': month[~mask_missing], 
                                                              'day': day[~mask_missing]}))
    return date_series

def back1year(date_series):
    return date_series - pd.DateOffset(years=1)






def namax(series_list):
    df = pd.concat(series_list, axis=1).convert_dtypes()
    return df.max(axis=1, skipna=True)

def namin(series_list):
    df = pd.concat(series_list, axis=1).convert_dtypes()
    return df.min(axis=1, skipna=True)

def namean(series_list):
    df = pd.concat(series_list, axis=1).convert_dtypes()
    return df.mean(axis=1, skipna=True)

def nasum(data: List[pd.Series]):

    use_df = pd.concat(data, axis=1)
    sum_series = use_df.sum(axis=1)
    sum_series[(sum_series == 0) & use_df.isnull().all(axis=1)] = pd.NA

    return sum_series

def leftsum(series_list):
    df = pd.concat(series_list, axis=1)

    result = df.sum(axis=1)

    result[df.iloc[:, 0].isnull()] = pd.NA

    return result

def naprod(series_list):
    df = pd.concat(series_list, axis=1).convert_dtypes()
    print(df)
    return df.product(axis=1, skipna=False)

def where(condition, x, y):
    if pd.isna(condition).any():  # check if there are any missing values
        return np.where(condition.fillna(False), x, y)
    else:
        return np.where(condition, x, y)
    

def replace(series, mapping_dict):

    if "NA" in mapping_dict:
        series = series.fillna(mapping_dict["NA"])

    mapping_dict.pop("NA", None)

    cross_type_replace = False
    same_type_replace_dict = {}
    cross_type_replace_dict = {}

    for k, v in mapping_dict.items():
        if isinstance(k, tuple):
            if any(type(elem) != type(v) for elem in k):
                cross_type_replace = True
                cross_type_replace_dict[k] = v
            else:
                same_type_replace_dict[k] = v
        else:
            if type(k) != type(v):
                cross_type_replace = True
                cross_type_replace_dict[k] = v
            else:
                same_type_replace_dict[k] = v

    if cross_type_replace:
        series = series.astype('object')
        series = series.replace(cross_type_replace_dict)

    series = series.replace(same_type_replace_dict)
    
    return series


def fill(target_s, source_s):
    """
    使用 source_s 来填充 target_s 中的缺失值。

    参数:
        target_s (pd.Series): 需要填充缺失值的 Series
        source_s (pd.Series): 用来填充缺失值的 Series

    返回:
        pd.Series: 更新后的 Series
    """
    return target_s.fillna(source_s)


def nalog(arr):
    # 将pd.NA转换为np.nan
    arr = arr.replace({pd.NA: np.nan})


    # 尝试计算对数
    try:
        result = arr.apply(np.log)
    except ValueError as err:
        # 打印错误信息和无效的输入值
        print(f"Error: {err}")

    return result
    


    


    
# transform函数
def transformer(func: Callable):
    @functools.wraps(func)
    def wrapper(df: pd.DataFrame, *args, **kwargs) -> pd.DataFrame:
        groupby = kwargs.pop('groupby', None)
        ascendby = kwargs.pop('ascendby', None)
        gascendby = kwargs.pop('gascendby', None)
        to_nullable_dtypes = kwargs.pop('to_nullable_dtypes', False)
        columns = kwargs.pop('columns',df.columns.values)
        

        if groupby and ascendby and groupby == ascendby:
            raise ValueError("groupby and ascendby cannot have the same value")

        df = df.copy()

        if gascendby:
            df.sort_values(by=gascendby, kind='mergesort', inplace=True) # 有些数据类型如混合类型无法排序
            
        if groupby:
            if ascendby:
                df.sort_values(by=ascendby, kind='mergesort', inplace=True) # 有些数据类型如混合类型无法排序
            df[columns] = df.groupby(groupby)[columns].transform(lambda x: func(x, *args, **kwargs))
        else:
            df[columns] = df[columns].transform(lambda x: func(x, *args, **kwargs))

        if to_nullable_dtypes:
            df[columns] = df[columns].convert_dtypes()
        return df
    return wrapper


## 以下除装饰器函数外，均为df.transform可以使用的关于pd.Series的函数

def boardcast_to_series(include_dtypes: Literal['number','all','string','category'], to_numpy_nan: bool = False):
    def decorator(func: Callable):
        @functools.wraps(func)  
        def wrapper(s: pd.Series, *args, **kwargs):
            dtype_map = {
                'number': is_numeric_dtype,
                'string': is_string_dtype,
                'category': is_categorical_dtype,
                'all': pandas_dtype
            }
            
            if dtype_map[include_dtypes](s.dtype):
                if to_numpy_nan:
                    s = s.fillna(np.nan)
                return pd.Series(np.repeat(func(s, *args, **kwargs), len(s)), index=s.index)
            else:
                return s
        return wrapper
    return decorator


# 非缺失数值统计
@transformer
@boardcast_to_series(include_dtypes='number', to_numpy_nan= True)
def cmean(s: pd.Series):
    return s.mean()

@transformer
@boardcast_to_series(include_dtypes='number', to_numpy_nan= True)
def cmin(s: pd.Series):
    return s.min()

@transformer
@boardcast_to_series(include_dtypes='number', to_numpy_nan= True)
def cmax(s: pd.Series):
    return s.max()


@transformer
@boardcast_to_series(include_dtypes='number', to_numpy_nan= True)
def cstd(s: pd.Series):
    return s.std()

@transformer
@boardcast_to_series(include_dtypes='number', to_numpy_nan= True)
def csum(s: pd.Series):
    return s.sum()

@transformer
@boardcast_to_series(include_dtypes='number', to_numpy_nan= True)
def cmedian(s: pd.Series):
    return s.median()

@transformer
@boardcast_to_series(include_dtypes='number', to_numpy_nan= True)
def cvar(s: pd.Series):
    return s.var()

# 独特非缺失值统计
@transformer
@boardcast_to_series(include_dtypes='all')
def cnunique(s: pd.Series):
    return s.nunique()

# 缺失值统计
@transformer
@boardcast_to_series(include_dtypes='all')
def cfirst_notna(s: pd.Series):
    index = s.first_valid_index()
    return s.loc[index] if index else np.nan

@transformer
@boardcast_to_series(include_dtypes='all')
def clast_notna(s: pd.Series):
    index = s.last_valid_index()
    return s.loc[index] if index else np.nan

@transformer
@boardcast_to_series(include_dtypes='all')
def ccount_notna(s: pd.Series):
    return s.count()

@transformer
@boardcast_to_series(include_dtypes='all')
def ccount_all(s: pd.Series):
    return s.size

@transformer
@boardcast_to_series(include_dtypes='all')
def ccount_na(s: pd.Series):
    return s.size-s.count()

@transformer
@boardcast_to_series(include_dtypes='all')
def ccount_na_pct(s: pd.Series):
    return (s.size-s.count())/s.size

@transformer
@boardcast_to_series(include_dtypes='all')
def ciloc(s: pd.Series, index: int):
    return s.iloc[index]

@transformer
def cffillna(s:pd.Series):
    return s.ffill()

@transformer
def cbfillna(s:pd.Series):
    return s.bfill()

# 以下除装饰器函数外，均为对df应该transform操作的函数， df->df
## 装饰器
def reval(df: pd.DataFrame, exprs: List[str]):
    
    expr = '\n'.join(exprs)
    return df.eval(expr)



def rassign(df: pd.DataFrame, exprs: Dict[str, str]):
    lambda_exprs = {k: eval('lambda x: ' + v) for k, v in exprs.items()}
    return df.assign(**lambda_exprs)

    
